package com.me.brickbuster.entity;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.glutils.ShapeRenderer;
import com.badlogic.gdx.graphics.glutils.ShapeRenderer.ShapeType;
import com.badlogic.gdx.math.MathUtils;
import com.badlogic.gdx.math.Vector2;
import com.me.brickbuster.BrickBuster;
import com.me.brickbuster.Utils;
import com.me.brickbuster.state.PlayState;
import net.dermetfan.utils.Pair;

import java.util.Iterator;

public class Ball extends Entity {

	public static final int RADIUS = 12;
	public static final Color BALL_COLOR = Color.CHARTREUSE;
	public static final float DEFAULT_SPEED = 350;
	public static final float BOOST_SPEED = 450;
	public static final int BLOCKS_FOR_BOOST = 39;

	public Vector2 direction;
	private float speed;
	private boolean isStuck = true;
	private boolean isDead = false;

	public Ball(PlayState state) {
		super(state,BrickBuster.WIDTH/2, state.paddle.getY() + Paddle.PADDLE_HEIGHT + RADIUS);
		this.speed = state.bricks.size() > BLOCKS_FOR_BOOST? DEFAULT_SPEED : BOOST_SPEED;
	}

	@Override
	public void render(ShapeRenderer sr) {
		sr.begin(ShapeType.Filled);
		sr.setColor(BALL_COLOR);
		sr.circle(pos.x, pos.y, RADIUS);
		sr.end();
	}

	@Override
	public void update(float dt) {
		if (isStuck || isDead) {
			if (!isDead && Gdx.input.justTouched()) {
				launch();
			} else {
				return;
			}
		}

		Vector2 new_pos = pos.cpy().add(direction.cpy().scl(speed * dt));

		boolean brickCollision = false;
		Iterator<Brick> brickIterator = state.bricks.iterator();
		while (!brickCollision && brickIterator.hasNext()) {
			Brick brick = brickIterator.next();
			Vector2[] vertices = brick.getVertices();
			for(int i = 0; i < vertices.length; i++) {
				Vector2 v1 = vertices[i];
				Vector2 v2 = vertices[i+1 < vertices.length? i+1 : 0];
				Vector2 segment = v2.cpy().sub(v1);
				Vector2 nearest = Utils.nearestPoint(v1.cpy(), segment, new_pos.cpy());

				if (nearest.dst(new_pos.x, new_pos.y) <= RADIUS) {
					if (brick.hit()) {
						brickIterator.remove();
					}
					Utils.reflect(direction, segment.nor());
					brickCollision = true;
					break;
				}
			}
		}

		if (new_pos.x + RADIUS > BrickBuster.WIDTH || new_pos.x - RADIUS < 0) {
			Utils.reflect(direction, Utils.VERTICAL_EDGE);
		} else if (new_pos.y + RADIUS > BrickBuster.HEIGHT) {
			Utils.reflect(direction, Utils.HORIZONTAL_EDGE);
		} else if (state.getShieldCount() > 0
					&& new_pos.y - RADIUS < PlayState.SHIELD_HEIGHT * state.getShieldCount()) {
			Utils.reflect(direction, Utils.HORIZONTAL_EDGE);
			state.removeShield();
		} else if (new_pos.y + RADIUS < 0) {
			isDead = true;
			return;
		} else if (direction.y < 0 && new_pos.y <= state.paddle.getY() + Paddle.PADDLE_HEIGHT + RADIUS) {
			Pair<Vector2, Vector2> paddle = state.paddle.getTopEdge();
			Vector2 lineDir = paddle.getValue().sub(paddle.getKey());
			Vector2 nearest = Utils.nearestPoint(paddle.getKey().cpy(), lineDir, new_pos.cpy());

			if (nearest.dst(new_pos.x, new_pos.y) <= RADIUS) {
				paddleCollision();
				if (state.paddle.isSticky()) {
					return;
				}
			}
		}

		pos.add(direction.cpy().scl(speed * dt));
	}

	public Vector2 paddleReflectAngle() {
		float rel = MathUtils.clamp((pos.x - state.paddle.getX()) + (state.paddle.getWidth()/2), 5, state.paddle.getWidth()-5);
		float newAngle = MathUtils.PI - (MathUtils.PI * (rel / state.paddle.getWidth()));
		return new Vector2(MathUtils.cos(newAngle), MathUtils.sin(newAngle));
	}

	public void launch() {
		if (state.paddle.isSticky()) {
			direction = paddleReflectAngle();
		} else {
			// launch at random angle between 135 and 45 degrees
			float angle = MathUtils.random(MathUtils.PI/2) + MathUtils.PI/4;
			direction = new Vector2(MathUtils.cos(angle), MathUtils.sin(angle));
		}
		isStuck = false;
	}

	public void paddleCollision() {
		if (state.paddle.isSticky()) {
			isStuck = true;
			pos.y = state.paddle.getY() + Paddle.PADDLE_HEIGHT + RADIUS;
			return;
		}
		direction = paddleReflectAngle();
	}

	public void setSpeed(float speed) {
		this.speed = speed;
	}

	public boolean isStuck() {
		return isStuck;
	}

	public boolean isDead() {
		return isDead;
	}

	public void setDirection(Vector2 direction) {
		this.direction = direction;
	}

	public void setStuck(boolean stuck) {
		isStuck = stuck;
	}
}
